import logging
import socket
import sys
import threading
from optparse import OptionParser
from subprocess import PIPE, Popen

errorList = []


# Update these locations as per requirements
# Default log location will fetch everything and might increase support bundle size
LINUX_LOG_LOCATIONS = "/var/log/pods, /var/log/containers, /var/log/calico, /var/lib/docker, /var/lib/nfs, /var/lib/kubelet"
LINUX_CONFIG_LOCATIONS = "/etc/docker, /etc/kubernetes"
WINDOWS_CYGWIN_ROOT = "/cygdrive/c"
WINDOWS_LOG_LOCATIONS = "/var/log/pods, /var/log/containers, /var/log/kubelet, /var/lib/kubelet, /var/lib/cni, /var/lib/kube-proxy, /etc/kubernetes/logs"
WINDOWS_CONFIG_LOCATIONS = "/etc/rancher, /etc/kubernetes, /etc/cni"

# System namespace objects
NAMESPACES_SYSTEM = "kube-system, vmware-system-csi, cns-manager"
RESOURCES_SYSTEM = "pod, statefulset, deployment, daemonset, configmap, secret"

# User namespaces objects. Currently just default is being considered as user namespaces
NAMESPACES_USER = ""
RESOURCES_USER = "pvc, pod, statefulset, deployment"

# Cluster objects
RESOURCES_CLUSTER = "node, pv, volumeattachment"


def GetOptions():
    """
    Supports the command-line arguments listed below
    """
    if "win32" in sys.platform:
        tmpDir = r"C:\Users\Administrator\AppData\Local\Temp\\"
    else:
        tmpDir = "/tmp/"

    parser = OptionParser()
    parser.add_option(
        "-d",
        "--dest",
        default=tmpDir,
        help="Path to store log files [default: %default]",
    )
    parser.add_option("-t", "--nodeType", default="", help="Node type in use")
    parser.add_option(
        "-u",
        "--homeUser",
        default="ubuntu",
        help="Non-root user having access to cluster kubeconfig",
    )
    (options, _) = parser.parse_args()
    return options


def execShellCommand(command):
    """
    Executes command on localhost
    """
    try:
        process = Popen(args=command, stdout=PIPE, stderr=PIPE, shell=True)
        stdout, stderr = process.communicate()
        stdout = str(stdout, "utf-8")
        stderr = str(stderr, "utf-8")
        return stdout, stderr
    except Exception as e:
        print("Failed to execute command '%s'. Error: %s", command, str(e))


def make_tarfile(output_filename, source_dirs):
    """
    Creates tar file for provided directory tree
    """
    if "win32" in sys.platform:
        output_filename = WINDOWS_CYGWIN_ROOT + str(output_filename).replace(
            "\\", "/"
        ).replace("C:", "")
        TAR_COMMAND = r"C:\cygwin64\bin\bash -login -c 'tar --force-local --ignore-failed-read --mode=777 --exclude=/cygdrive/c/var/lib/docker/overlay2 -iczhf {} ".format(
            output_filename
        )
    else:
        TAR_COMMAND = "tar --ignore-failed-read --exclude='/var/lib/docker/overlay2' -iczhf {} ".format(
            output_filename
        )
    for directory in source_dirs:
        directory = directory.strip()
        if "win32" in sys.platform:
            directory = WINDOWS_CYGWIN_ROOT + str(directory).replace("\\", "/").replace(
                "C:", ""
            )
        print("Collecting logs from {} ... \n".format(directory))
        # with tarfile.open(output_filename, "w:gz") as tar:
        #     tar.add(directory, arcname=os.path.basename(basename))
        if output_filename == directory[:-1] + ".tgz":
            TAR_COMMAND = TAR_COMMAND + " -C {} . ".format(directory)
        else:
            TAR_COMMAND = TAR_COMMAND + directory + " "

    if "win32" in sys.platform:
        TAR_COMMAND = TAR_COMMAND + "'"
    print(TAR_COMMAND)
    _, stderr = execShellCommand(str(TAR_COMMAND))
    if stderr != "":
        errorList.append(
            "Error creating tar file for support bundle: {}".format(stderr.strip())
        )


def createSupportDirectory():
    if "win32" in sys.platform:
        separator = "\\"
        mkdirCmd = "mkdir "
    else:
        separator = "/"
        mkdirCmd = "mkdir -p "
    global k8s_support_dir_path
    k8s_support_dir_name = str(socket.gethostname())
    _, stderr = execShellCommand(str(mkdirCmd + options.dest + k8s_support_dir_name))
    k8s_support_dir_path = str(options.dest + k8s_support_dir_name + separator)
    if stderr != "":
        errorList.append("Error creating support bundle: {}".format(stderr))
    return k8s_support_dir_name


def collectKubeletLogs():
    """
    Collects kubelet Logs from all nodes(master + workers) in k8s cluster and stores in sub-directory kubelet
    """
    global kubeletLogsDestDirectoryPath
    kubeletLogsDestDirectoryPath = str(k8s_support_dir_path + "kubelet/")

    try:
        # Create directory to collect kubelet logs
        execShellCommand(str("mkdir -p " + kubeletLogsDestDirectoryPath))

        # Retain only last 10 days of journal logs
        execShellCommand(str("sudo journalctl --vaccum-time=10d"))

        kubeletLogCommand = str(
            "journalctl -xeu kubelet > "
            + kubeletLogsDestDirectoryPath
            + options.nodeType
            + "-"
            + str(socket.gethostname())
            + "-kubelet.log"
        )
        execShellCommand(kubeletLogCommand)
    except Exception as e:
        print("Failed to collect kubelet logs due to :{}".format(str(e)))
        log.error("Failed to collect kubelet logs. Error : {}".format(str(e)))


def runNamespaceResourceDescribe(resourceObj, resourceNamespace, outputPath):
    """
    Runs kubectl describe on provided rescource and namespace in required output path
    """
    try:
        command = KUBECTL_CMD + "get {} -n {} | awk '{{print $1}}' | tail -n +2".format(
            resourceObj, resourceNamespace
        )
        stdout_resources, stderr_resources = execShellCommand(command)
        if stderr_resources != "":
            if "No" in str(stderr_resources):
                errorList.append(
                    "Error getting "
                    + resourceObj
                    + " names: {}".format(stderr_resources.strip())
                )
                return None
        resources = stdout_resources.split("\n")
    except Exception as e:
        print("Failed to get resources. Error: {}".format(str(e)))
        return None

    for res in resources:
        try:
            RESOURCE_DESCRIBE_COMMAND = (
                KUBECTL_CMD
                + "describe {} {} -n {} >> {}_describe.txt".format(
                    resourceObj, res, resourceNamespace, outputPath + resourceObj
                )
            )
            _, stderr_resource_describe = execShellCommand(RESOURCE_DESCRIBE_COMMAND)
            if stderr_resource_describe != "":
                errorList.append(
                    "Error describing " + res + " : {}".format(stderr_resource_describe)
                )
        except Exception as e:
            print(
                "Failed to get describe of %s %s. Error: %s", res, resourceObj, str(e)
            )


def runClusterResourceDescribe(resourceObj, outputPath):
    """
    Runs kubectl describe on provided cluster rescource in required output path
    """
    try:
        command = KUBECTL_CMD + "get {} | awk '{{print $1}}' | tail -n +2".format(
            resourceObj
        )
        stdout_resources, stderr_resources = execShellCommand(command)
        if stderr_resources != "":
            if "No" in str(stderr_resources):
                errorList.append(
                    "Error getting "
                    + resourceObj
                    + " names: {}".format(stderr_resources.strip())
                )
                return None
        resources = stdout_resources.split("\n")
    except Exception as e:
        print("Failed to get resources. Error: {}".format(str(e)))
        return None

    for res in resources:
        try:
            RESOURCE_DESCRIBE_COMMAND = (
                KUBECTL_CMD
                + "describe {} {} >> {}_describe.txt".format(
                    resourceObj, res, outputPath + resourceObj
                )
            )
            _, stderr_resource_describe = execShellCommand(RESOURCE_DESCRIBE_COMMAND)
            if stderr_resource_describe != "":
                errorList.append(
                    "Error describing " + res + " : {}".format(stderr_resource_describe)
                )
        except Exception as e:
            print(
                "Failed to get describe of %s %s. Error: %s", res, resourceObj, str(e)
            )


def getObjectEvents(objectType, objectOutputPath, resourceTypes):
    """
    Collects cluster resource describe events and stores at sub-directory describe_events
    """
    if "win32" in sys.platform:
        separator = "\\"
        mkdirCmd = "mkdir -Force "
    else:
        separator = "/"
        mkdirCmd = "mkdir -p "
    try:
        commandsDestDirectoryPath = str(
            objectOutputPath + "describe_events" + separator
        )

        # Create directory to collect cluster resource events
        execShellCommand(str(mkdirCmd + commandsDestDirectoryPath))
    except Exception as e:
        print(
            "Failed to create support sub-directory describe_events. Error: %s", str(e)
        )
        return None

    for res in resourceTypes.split(","):
        res = res.strip()
        print(
            "Collecting describe events of all {} in {} ...  \n".format(res, objectType)
        )

        # Collect resource describe events
        if objectType == "cluster":
            runClusterResourceDescribe(res, commandsDestDirectoryPath)
        else:
            runNamespaceResourceDescribe(res, objectType, commandsDestDirectoryPath)


def runClusterCommands():
    """
    Runs various commands to collect cluster Information and stores in sub-directory commands
    """
    try:
        if "win32" in sys.platform:
            separator = "\\"
            mkdirCmd = "mkdir -Force "
        else:
            separator = "/"
            mkdirCmd = "mkdir -p "
        commandsDestDirectoryPath = str(k8s_support_dir_path + "commands" + separator)

        # Create directory to collect cluster command outputs
        execShellCommand(str(mkdirCmd + commandsDestDirectoryPath))

        # Create directory to collect cluster objects outputs
        clusterOutpuDestDirectoryPath = str(
            commandsDestDirectoryPath + "cluster" + separator
        )
        execShellCommand(str(mkdirCmd + clusterOutpuDestDirectoryPath))

        # Collect resource info of cluster objects
        for res in RESOURCES_CLUSTER.split(","):
            res = res.strip()
            print("Collecting {} cluster resource info ... \n".format(res))

            # Collects detailed info on cluster resource
            kubeSystemPodInfoCommand = str(
                "{} get {} -o yaml > {}{}.yaml".format(
                    KUBECTL_CMD, res, clusterOutpuDestDirectoryPath, res
                )
            )
            execShellCommand(kubeSystemPodInfoCommand)

            # Collects minimum required info on all system pods in a cluster
            kubeSystemPodInfoCommand = str(
                "{} get {} -o wide > {}{}.min.txt".format(
                    KUBECTL_CMD, res, clusterOutpuDestDirectoryPath, res
                )
            )
            execShellCommand(kubeSystemPodInfoCommand)

        # Collect describe events for namespace objects
        getObjectEvents("cluster", clusterOutpuDestDirectoryPath, RESOURCES_CLUSTER)

        for ns in NAMESPACES_SYSTEM.split(","):
            ns = ns.strip()
            # Create directory to collect system namespace outputs
            namespaceDestDirectoryPath = str(commandsDestDirectoryPath + ns + separator)
            execShellCommand(str(mkdirCmd + namespaceDestDirectoryPath))

            # Collect resource info in current namespace context
            for res in RESOURCES_SYSTEM.split(","):
                res = res.strip()
                print(
                    "Collecting {} resource info from {} namespace ... \n".format(
                        res, ns
                    )
                )

                # Collects detailed info on all system pods in a cluster
                kubeSystemPodInfoCommand = str(
                    "{} get {} -n {} -o yaml > {}{}.yaml".format(
                        KUBECTL_CMD, res, ns, namespaceDestDirectoryPath, res
                    )
                )
                execShellCommand(kubeSystemPodInfoCommand)

                # Collects minimum required info on all system pods in a cluster
                kubeSystemPodInfoCommand = str(
                    "{} get {} -n {} -o wide > {}{}.min.txt".format(
                        KUBECTL_CMD, res, ns, namespaceDestDirectoryPath, res
                    )
                )
                execShellCommand(kubeSystemPodInfoCommand)

            # Collect describe events for namespace objects
            getObjectEvents(ns, namespaceDestDirectoryPath, RESOURCES_SYSTEM)

        default_user_namespace = "default"
        # Create directory to collect system namespace outputs
        namespaceDestDirectoryPath = str(
            commandsDestDirectoryPath + default_user_namespace + separator
        )
        execShellCommand(str(mkdirCmd + namespaceDestDirectoryPath))

        for res in RESOURCES_USER.split(","):
            res = res.strip()
            print(
                "Collecting {} resource info from {} namespace ... \n".format(
                    res, default_user_namespace
                )
            )

            # Collects detailed info of resources in cluster
            resInfoCommand = str(
                "{} get {} -o yaml > {}{}.yaml".format(
                    KUBECTL_CMD, res, namespaceDestDirectoryPath, res
                )
            )
            execShellCommand(resInfoCommand)

            # Collects minimum required info of resources in a cluster
            resInfoCommand = str(
                "{} get {} -o wide > {}{}.min.txt".format(
                    KUBECTL_CMD, res, namespaceDestDirectoryPath, res
                )
            )
            execShellCommand(resInfoCommand)

        # Collect describe events for namespace objects
        getObjectEvents(
            default_user_namespace, namespaceDestDirectoryPath, RESOURCES_USER
        )
    except Exception as e:
        print("Failed to run cluster commands on k8s clusters. Error: %s", str(e))


def provideFilePermissions(directory):
    CHMOD_COMMAND = str("chmod -R 755 " + directory)
    execShellCommand(CHMOD_COMMAND)
    CHOWN_COMMAND = str("chown -R " + options.homeUser + " " + directory)
    execShellCommand(CHOWN_COMMAND)


def main():
    # Process Command line args
    global options
    options = GetOptions()

    # Create Support directory
    global support_dir

    try:
        support_dir = createSupportDirectory()
    except Exception as e:
        print("Failed to Create Support Directory due to : %s", str(e))
        log.error("Failed to create support dirctory. Error : %s", str(e))
        return

    # Prepare kubeconfig path
    global KUBECTL_CMD
    KUBECTL_CMD = str(
        "kubectl --kubeconfig=/home/'{}'/.kube/config ".format(options.homeUser)
    )

    # Collect Cluster Info
    if "worker" not in options.nodeType:
        runClusterCommands()

    # Collect kubelet logs
    if "win32" not in sys.platform:
        collectKubeletLogs()

    k8s_support_dir_paths = []
    k8s_support_dir_paths.append(k8s_support_dir_path)
    if "win32" in sys.platform:
        k8s_support_dir_paths.extend(WINDOWS_LOG_LOCATIONS.split(","))
        k8s_support_dir_paths.extend(WINDOWS_CONFIG_LOCATIONS.split(","))
    else:
        k8s_support_dir_paths.extend(LINUX_LOG_LOCATIONS.split(","))
        k8s_support_dir_paths.extend(LINUX_CONFIG_LOCATIONS.split(","))

    # Create compressed archive of support bundle
    try:
        generate_support_bundle_process = threading.Thread(
            name="support-bundle-create",
            target=make_tarfile(
                options.dest + support_dir + ".tgz", k8s_support_dir_paths
            ),
        )
        generate_support_bundle_process.start()
        screen_info = "K8s Cluster Support bundle '{}' generated at '{}'"
        screen_info_log = str(screen_info.format(support_dir + ".tgz", options.dest))
        while generate_support_bundle_process.is_alive():
            continue
        print(screen_info_log)
        log.info(screen_info_log)

        # Give Permissions to support bundle files
        provideFilePermissions(options.dest)
    except Exception as e:
        return


# Start script
if __name__ == "__main__":
    log = logging.getLogger(__name__)
    support_process = threading.Thread(name="k8s-support", target=main)
    support_process.start()
    while support_process.is_alive():
        continue
    print("\nComplete..")
    if len(errorList) > 0:
        print("Errors during script exec: \n", errorList)
